from typing import List, Union

from pydantic import BaseModel

from tktl.core.t import InstanceTypeT, ProjectAssetSourceT, ProjectAssetT, ServiceT


class ProjectContentsBase(BaseModel):
    type: str
    name: str
    path: str


class ProjectFile(ProjectContentsBase):
    type: str = "file"


class ProjectFileWithContent(ProjectContentsBase):
    type: str = "file"
    content: str


class ProjectDirectory(ProjectContentsBase):
    type: str = "dir"


class ProjectAsset(ProjectContentsBase):
    calculated_sha: str
    kind: ProjectAssetT
    source: ProjectAssetSourceT
    requires_download: bool


class TktlServiceConfigSchema(BaseModel):
    replicas: int
    instance_type: InstanceTypeT
    service_type: ServiceT


class TktlAssetConfigSchema(BaseModel):
    path: str
    version: int


class TktlYamlConfigSchema(BaseModel):

    """
    default_branch_name: master
    service:

      replicas: 2
      instance_type: gp.small
      service_type: rest

    model:
      path: assets/model.pkl
      version: 1

    data:
      path: assets/data.pkl
      version: 1

    """

    default_branch_name: str
    service: TktlServiceConfigSchema
    model: TktlAssetConfigSchema
    data: TktlAssetConfigSchema


class TktlYamlConfigValidationError(BaseModel):
    loc: str
    msg: str


class ProjectValidationOutput(BaseModel):
    title: str
    summary: str
    text: str


ProjectContentSingleItemT = Union[ProjectFile, ProjectDirectory]
ProjectContentMultiItemT = List[ProjectContentSingleItemT]
ProjectContentT = Union[
    ProjectContentSingleItemT, ProjectContentMultiItemT, ProjectFileWithContent
]



from typing import Callable, List

import numpy as np
import pandas as pd
import pytest
from pydantic import BaseModel

from tktl import Tktl
from tktl.core.exceptions.exceptions import ValidationException
from tktl.registration.validation import (
    get_input_and_output_for_profiling,
    input_pandas_representation,
    output_pandas_representation,
)


@pytest.mark.parametrize(
    "X,y",
    [
        (pd.DataFrame({"a": [1, 2, 3]}), pd.Series([4, 5, 5])),  # pandas
        (np.array([[1], [2], [3]]), np.array([4, 5, 5])),  # numpy
        ({"a": [1, 2, 3]}, [4, 5, 5]),  # base
    ],
)
def test_creation(X, y):
    tktl = Tktl()

    @tktl.endpoint(X=X, y=y, kind="tabular")
    def predict(X):
        return [0] * 3

    endpoint = tktl.endpoints[0]
    assert endpoint.input_schema.pandas_convertible is True
    assert endpoint.output_schema.pandas_convertible is True
    pred = endpoint.func(pd.DataFrame(X))
    assert pred == [0, 0, 0]


@pytest.mark.parametrize("kind", ["tabular", "regression", "binary"])
def test_tabular_kinds(kind):
    tktl = Tktl()
    X = pd.DataFrame({"a": [1, 2, 3]})
    y = pd.Series([4, 5, 5])

    @tktl.endpoint(X=X, y=y, kind=kind)
    def predict(X):
        return [0] * len(X)

    endpoint = tktl.endpoints[0]
    assert isinstance(endpoint.input_schema.value, pd.DataFrame)
    assert isinstance(endpoint.output_schema.value, pd.Series)
    pred = endpoint.func(pd.DataFrame(X))
    assert pred == [0, 0, 0]


def test_custom_kind():
    tktl = Tktl()

    with pytest.raises(ValueError):
        # Must define either payload or X, y
        @tktl.endpoint(kind="custom")
        def predict_untyped(x):
            return x

    class Payload(BaseModel):
        a: List

    class Response(BaseModel):
        a: int
        b: float
        c: List

    @tktl.endpoint(kind="custom", payload_model=Payload, response_model=Response)
    def predict_typed(x):
        return x

    for endpoint in tktl.endpoints:
        assert isinstance(endpoint.func, Callable)
        assert endpoint.func(["foo"]) == ["foo"]

    x = pd.DataFrame([{"a": 1, "b": 2}, {"a": 1, "b": 2}, {"a": 1, "b": 2}])
    y = pd.Series([1, 2, 3])

    @tktl.endpoint(kind="custom", X=x, y=y)
    def predict_typed(x):
        return x

    schema = tktl.endpoints[1].input_schema.pydantic.schema()
    assert schema["title"] == "DataFrame__predict_typed__input"
    assert schema["example"] == [{"a": 1, "b": 2}]


def test_func_shape():
    tktl = Tktl()
    X = {"a": [1, 2, 3]}
    y = [4, 5, 5]

    @tktl.endpoint(X=X, y=y)
    def predict(X):
        return [0] * (len(X) - 1)

    assert tktl.endpoints[0].profiling_supported is False


def test_input_shape():
    tktl = Tktl()
    X = {"a": [1, 2, 3]}
    y = [4, 5]

    @tktl.endpoint(X=X, y=y)
    def predict(X):
        return [0] * len(X)

    assert tktl.endpoints[0].profiling_supported is False


def test_validate_func():
    tktl = Tktl()

    X = {"a": [1, 2, 3]}
    y = [4, 5, 5]

    @tktl.endpoint(X=X, y=y, kind="regression")
    def predict_regression(X):
        return ["a", "b", "c"]  # must be numeric

    assert tktl.endpoints[0].profiling_supported is False

    tktl = Tktl()
    X = {"a": [1, 2, 3]}
    y = [4, 5, 5]

    with pytest.raises(ValidationException):

        @tktl.endpoint(X=X, y=y, kind="binary")
        def predict_binary(X):
            return [-2, 0, 1.5]  # must be in [0, 1]

        _ = tktl.endpoints[0].profiling_supported


def test_unknown_kind():
    with pytest.raises(ValidationException):
        tktl = Tktl()
        X = {"a": [1, 2, 3]}
        y = [4, 5, 5]

        @tktl.endpoint(X=X, y=y, kind="unknown kind")
        def predict(X):
            return [0] * len(X)


def test_missing():
    tktl = Tktl()
    X = pd.DataFrame({"a": [1, 2, 3]})
    y = pd.Series([0.1, 0.2, None])

    @tktl.endpoint(X=X, y=y, kind="regression")
    def predict(X):
        return [0] * len(X)

    with pytest.warns(UserWarning):
        x_raw = input_pandas_representation(
            value=tktl.endpoints[0].input_schema.value, reset_index=False
        )
        y_raw = output_pandas_representation(
            value=tktl.endpoints[0].output_schema.value, reset_index=False
        )
        get_input_and_output_for_profiling(x=x_raw, y=y_raw)


def test_large():
    tktl = Tktl()
    n = int(1e6) + 1
    X = pd.DataFrame({"a": np.random.uniform(size=n)})
    y = pd.Series(np.random.uniform(size=n))

    with pytest.warns(UserWarning):

        @tktl.endpoint(X=X, y=y, kind="regression")
        def predict(X):
            return [0] * len(X)


def test_custom():
    tktl = Tktl()

    class Payload(BaseModel):
        a: str

    class Response(BaseModel):
        a: str

    @tktl.endpoint(kind="custom", payload_model=Payload, response_model=Response)
    def string_endpoint(x):
        return x



from typing import Union

from tktl.core import ExtendedEnum
from tktl.core.schemas.repository import Endpoint, Repository, RepositoryDeployment


class UserRepoConfigFileNameT(str, ExtendedEnum):
    YAML = "tktl.yaml"
    YML = "tktl.yml"


class UserProjectFileT(str, ExtendedEnum):
    FILE = "file"
    DIRECTORY = "dir"


class RequiredUserProjectPathsT(str, ExtendedEnum):
    SRC = "src"
    TKTL_YML = "tktl.yml"
    TKTL_YAML = "tktl.yaml"
    DOCKERFILE = ".buildfile"
    REQS = "requirements.txt"

    @classmethod
    def strictly_required_files(cls):
        return {cls.DOCKERFILE.value, cls.REQS.value}

    @classmethod
    def strictly_required_dirs(cls):
        return {cls.SRC.value}


class DeploymentStatesT(str, ExtendedEnum):
    PRE_ORION_UNKNOWN_ERROR = "errored"
    TEST_FAILED = "test_failed"
    IMAGE_BUILD_FAILED = "image_build_failed"
    DEPLOYMENT_FAILED = "deployment_failed"
    CANCELED = "canceled"
    SUCCEEDED = "succeeded"
    TEST_RUNNING = "test_running"
    IMAGE_BUILDING = "image_building"
    DEPLOYING = "deploying"
    MARKED_FOR_DELETION = "marked_for_deletion"
    DELETION_FAILED = "deletion_failed"


class ProjectAssetT(str, ExtendedEnum):
    DATA = "data"
    MODEL = "model"


class ProjectAssetSourceT(str, ExtendedEnum):
    S3 = "s3"
    LOCAL = "local"
    LFS = "lfs"


class InstanceTypeT(str, ExtendedEnum):
    GP_SMALL = "gp.small"
    GP_MEDIUM = "gp.medium"
    GP_LARGE = "gp.large"
    GP_XLARGE = "gp.xlarge"
    GP_2XLARGE = "gp.2xlarge"


class ServiceT(str, ExtendedEnum):
    REST = "rest"
    GRPC = "grpc"


class EndpointKinds(str, ExtendedEnum):
    CUSTOM = "custom"
    TABULAR = "tabular"
    BINARY = "binary"
    REGRESSION = "regression"
    MULTICLASS = "multiclass"


class RestSchemaTypes(ExtendedEnum):
    DICT = "Dict"
    SEQUENCE = "Sequence"
    ARRAY = "Array"
    FLAT_ARRAY = "FlatArray"
    SERIES = "Series"
    DATAFRAME = "DataFrame"
    SINGLE_VALUE = "SingleValue"
    CUSTOM_MODEL = "CustomModel"


Resources = Union[Repository, RepositoryDeployment, Endpoint]



TEMPLATE_PROJECT_DIRS = [
    "src/",
    "assets/",
    "tests/",
]

TEMPLATE_PROJECT_FILES = [
    "template/src/endpoints.py",
    "template/assets/model.joblib",
    "template/assets/loans_test.pqt",
    "template/tests/test_accuracy.py",
    "template/tests/test_plausibility.py",
    "template/tests/test_schema.py",
    "template/.gitignore",
    "template/.gitattributes",
    "template/.buildfile",
    "template/.dockerignore",
    "template/README.md",
    "template/requirements.txt",
]


CONFIG_FILE_TEMPLATE = """

# The default branch from which models will vbe deployed
default_branch_name: master

service:

  # The number of replicas of your service
  # Replicas are exact copies of your service, each request will be randomly directed to each of them
  replicas: 1

  # Format: <instance_kind>.<instance_size>
  # Only 'gp' (general purpose) instance_kind is supported for now
  # instance_size can be: small, medium, large
  instance_type: gp.small

  # Can be rest or grpc
  service_type: rest

# Assets: model and data
# Assets are the artifacts that comprise model and data.  These will be loaded inside the container
# when it gets built. Tktl will automatically package them for you, so all you need to do
# is define how they are loaded in the `src/endpoints.py file`


# Supported paths:
#  - A local path (default) which will use git-lfs to store it GitHub's server
#  - An s3 path. If this is picked and the bucket is private, the correct AWS keys should
#    be provided

# Versioning
# Every time someone pushes code to your repository, a deployment may be created. The deployment will
# only happen however if a new version of the model is found. That happens in three cases:

# 1. A different version is specified in this file
# 2. The md5 of the model or data has changed. This is detected automatically by taktile
# 3. The source code inside `src` has changed

# With events 2 and 3, a new version of the model will be published, and the tktl.yaml file will be
# automatically updated by the Taktile GitHub App

model:
  path: assets/model.pkl
  version: 1

data:
  path: assets/data.pkl
  version: 1

version: {version}
"""



from tktl.commands.validate import build_image
from tktl.core.exceptions.exceptions import MissingDocker
from tktl.core.loggers import LOG


def run_container(
    path: str, nocache: bool, background: False, auth_enabled: bool = True
):
    try:
        dm, image = build_image(path, nocache=nocache)
    except MissingDocker:
        LOG.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )
        return
    LOG.log("Waiting for service to start...")
    container = dm.run_container(image, detach=True, auth_enabled=auth_enabled)
    if background:
        return container
    try:
        dm.stream_logs(container)
    except KeyboardInterrupt:
        container.kill()
